﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Threading.Tasks;
using MongoDB.Driver;
using Robware.Blog;
using Robware.Blog.Data.MongoDB.State;

namespace Robware.Data.MongoDB {
	public class BlogRepository : IBlogRepository {
		private readonly IMongoCollection<BlogPostState> _collection;

		public BlogRepository(IMongoDatabase database) {
			_collection = database.GetCollection<BlogPostState>("blog");
		}

		private BlogPost MapStateToPost(BlogPostState state) =>
			new BlogPost {
				Id = state.Id,
				Title = state.Title,
				Content = state.Content,
				Timestamp = state.Timestamp,
				Draft = state.Draft,
				Url = state.Url,
				UserId = state.UserId,
			};

		private IEnumerable<BlogPost> MapStateToPost(IEnumerable<BlogPostState> states) => states.Select(MapStateToPost);

		public async Task DeletePostAsync(int id) {
			var update = Builders<BlogPostState>.Update.Set(nameof(BlogPostState.Deleted), true);
			await _collection.UpdateOneAsync(post => post.Id == id, update);
		}

		public async Task<IEnumerable<BlogPost>> GetAllPostsAsync() => MapStateToPost((await _collection.FindAsync(post => true)).ToEnumerable());

		public async Task<int> GetCountAsync() => (int)await _collection.CountDocumentsAsync(post => !post.Deleted);

		public async Task<BlogPost> GetLatestPostAsync() => (await GetLatestPostsAsync(1)).FirstOrDefault();

		public async Task<IEnumerable<BlogPost>> GetLatestPostsAsync(int limit, int offset = 0) {
			var filter = Builders<BlogPostState>.Filter.Eq(nameof(BlogPostState.Deleted), false);
			var sort = Builders<BlogPostState>.Sort.Descending(nameof(BlogPostState.Timestamp));
			var options = new FindOptions<BlogPostState> {
				Sort = sort,
				Limit = limit,
				Skip = offset
			};
			var states = (await _collection.FindAsync<BlogPostState>(filter, options)).ToEnumerable();
			return MapStateToPost(states);
		}

		private async Task<BlogPost> GetPostAsync(Expression<Func<BlogPostState, bool>> filter) {
			var result = (await _collection.FindAsync(filter)).FirstOrDefault();
			if (result == null)
				throw new ItemNotFoundException(nameof(GetPostAsync), null);
			return MapStateToPost(result);
		}

		public async Task<BlogPost> GetPostByIdAsync(int id) => await GetPostAsync(post => post.Id == id && !post.Deleted);

		public async Task<BlogPost> GetPostByUrlAsync(string url) => await GetPostAsync(post => post.Url == url && !post.Deleted);

		public async Task<BlogPost> SavePost(BlogPost post) {
			if (post.Id == 0) {
				post.Id = (await GetCountAsync()) + 1;
				post.Timestamp = DateTime.Now;
			}

			var mongoPost = new BlogPostState(post);
			await _collection.InsertOneAsync(mongoPost);

			return post;
		}
	}
}
